
import sys
import os

from .runner_factory import Runner, Factory

_SYSTEM_FACTORY = None

class PythonREPL(Runner):

    ICON = ('icons.flow', 'python')
    TAGS = ['Scripting']

    @classmethod
    def runner_name(cls):
        return 'Python'

    def keep_terminal(self):
        return False

    def executable(self):
        return sys.executable
    
    def argv(self):
        return []

class CMD(Runner):

    TAGS = ['Scripting']
    ICON = ('icons.flow', 'workstation')

    def show_terminal(self):
        return True

    def keep_terminal(self):
        return True

    def executable(self):
        return 'echo'
    
    def argv(self):
        return 'Here is a terminal, have fun!'.split()

class Explorer(Runner):

    ICON = ('icons.flow', 'explorer')
    TAGS = ['Open', 'Reveal']

    @classmethod
    def runner_name(cls):
        return 'Explorer'

    def show_terminal(self):
        return False

    def executable(self):
        return 'explorer.exe'
    
    def argv(self):
        return []


def get_system_factory():
    global _SYSTEM_FACTORY
    if _SYSTEM_FACTORY is None:
        factory = Factory('System')
        if os.name == 'nt':
            factory.ensure_runner_type(Explorer)
            factory.ensure_runner_type(CMD)
        factory.ensure_runner_type(PythonREPL)
        _SYSTEM_FACTORY = factory
    return _SYSTEM_FACTORY




import os

from kabaret.app._actor import Actor, Cmd, Cmds

from .. import runner_factory
from ..runners import get_system_factory

class SubprocessManagerCmds(Cmds):
    pass

@SubprocessManagerCmds.cmd
class List_Runners(Cmd):
    '''
    Returns an ordered list of dicts with keys:
        (
            factory_name, 
            group_name, 
            group_icon,
            runner_name, 
            runner_icon,
            runner_tags,
            version
        )
        NB: one entry per supported version !
    '''

    def _decode(self, reverse=False):
        self.reverse = reverse
    
    def _execute(self):
        collected = []
        factories = self.actor().get_factories()
        last_group_name = None
        for factory, RunnerType in factories.list_runner_types():
            factory_name = factory.name

            runner_name = RunnerType.runner_name()
            runner_icon = RunnerType.runner_icon()
            supported_versions = RunnerType.supported_versions()

            group_name = None            
            tags = RunnerType.runner_tags()
            if tags:
                group_name = tags[0]
            
            if last_group_name is None or group_name != last_group_name:
                last_group_name = group_name
                group_icon = runner_icon

            for version in supported_versions:
                sortable = [
                    factory_name, group_name, group_icon,
                    runner_name, runner_icon,
                ]
                if version is not None:
                    sortable.append(version)
                collected.append((
                    tuple(sortable), 
                    dict(
                        factory_name=factory.name,
                        group_name=group_name,
                        group_icon=group_icon,
                        runner_name=runner_name,
                        runner_icon=runner_icon,
                        runner_tags=tags,
                        version=version,
                    )
                ))
        collected.sort(reverse=self.reverse)
        return [i[1] for i in collected]

@SubprocessManagerCmds.cmd
class Get_Runner_Versions(Cmd):

    def _decode(
        self, 
        runner_name, tags=[]
    ):
        self.runner_name = runner_name
        self.tags = tags
    
    def _execute(self):
        return self.actor().get_factories().get_runner_versions(
            self.runner_name, self.tags
        )

@SubprocessManagerCmds.cmd
class Run(Cmd):

    def _decode(
        self, 
        runner_name, tags=[], version=None,
        label=None, extra_argv=[], extra_env={},
    ):
        self.runner_name = runner_name
        self.tags = tags
        self.version = version
        self.label = label
        self.extra_argv = extra_argv
        self.extra_env = extra_env

    def _execute(self):
        self.actor().run(
            self.runner_name, 
            self.tags,
            self.version,
            self.label,
            self.extra_argv,
            self.extra_env,
        )    

class SubprocessManager(Actor):
    
    def __init__(self, session):
        super(SubprocessManager, self).__init__(session)
        # self._manager = manager.SubprocessManager()
        self._factories = runner_factory.Factories()
        self._subprocess_manager = runner_factory.SubprocessManager()

        # Add default factories:
        self._factories.ensure_factory(
            get_system_factory()
        )

    def _create_cmds(self):
        return SubprocessManagerCmds(self)

    def get_factories(self):
        return self._factories

    def create_new_factory(self, name):
        return self._factories.create_new_factory(name)

    def ensure_factory(self, runner_factory):
        self._factories.ensure_factory(
            runner_factory
        )
        self.session().dispatch_event(
            'subprocess_manager.factory_added', 
            factory_name=runner_factory.name
        )

    def run(
            self, 
            runner_name, tags=[], 
            version=None, label=None,
            extra_argv=[], extra_env={},
        ):
        runner = self._factories.get_runner(
            runner_name, tags, 
            version, label,
            extra_argv, extra_env,
        )
        self._subprocess_manager.run(
            runner
        )



from kabaret.app import resources

import kabaret.app.ui.gui.widgets.widget_view
from kabaret.app.ui.gui.widgets.widget_view import (
    ToolBarView,
    QtWidgets, QtGui, QtCore
)


class LauncherToolBar(ToolBarView):

    def __init__(self, *args, **kwargs):
        kwargs['area'] = QtCore.Qt.LeftToolBarArea
        super(LauncherToolBar, self).__init__(*args, **kwargs)

        self.refresh()

    def refresh(self):
        self.clear()
        self.load()

    def clear(self):
        actions = self.actions()
        for action in actions:
            self.removeAction(action)
            action.deleteLater()

    def load(self):
        runners_list = self.session.cmds.SubprocessManager.list_runners(
            reverse=True
        )
        last_factory_name = None
        last_group_name = None
        tb = None
        for entry in runners_list:
            factory_name = entry['factory_name']
            group_name = entry['group_name']
            if last_factory_name != factory_name:
                self.addSeparator()
                last_factory_name = factory_name
            if tb is None or group_name != last_group_name:
                tb = self._create_group(entry)
                last_group_name = group_name
            self._add_runner(entry, tb)
            #self.ensure_group(factory_name, group_name)
    
    def _create_group(self, entry):
        
        factory_name = entry['factory_name']
        name = entry['group_name']
        icon_ref = entry['group_icon']
        
        tb = QtWidgets.QToolButton(self)
        self.addWidget(tb)
        
        tb.setProperty('hide_arrow', True)
        tb.setText(name)
        tb.setToolTip('{} ({})'.format(name, factory_name))
        tb.setPopupMode(tb.InstantPopup)
        tb.setIcon(resources.get_icon(icon_ref))
        tb.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        
        return tb

    def _add_runner(self, entry, tb):
        name = entry['runner_name']
        version = entry['version']
        tags = entry['runner_tags']
        icon = entry['runner_icon']

        label = name
        if version is not None:
            label = '{} ({})'.format(name, version)
        action = QtWidgets.QAction(tb)
        action.setText(label)
        action.setIcon(resources.get_icon(icon))
        action.triggered.connect(
            lambda n=name, t=tags, v=version: self._run(n, t, v)
        )
        action = tb.addAction(action)

    def _run(self, name, tags, version):
        self.session.cmds.SubprocessManager.run(
            name, tags, version
        )
        
    def receive_event(self, event_type, data):
        # print('!??? LauncerToolBar event', event_type, data)
        if event_type == 'subprocess_manager.factory_added':
            # TODO: we should sync instead of refresh all
            # But let's not optimize early ;)
            self.refresh()




import os
import subprocess

import logging
logger = logging.getLogger(__name__)

class Runner(object):
    '''
    The Runner is able to spawn a subprocess with controled
    command line and environment (among others things).

    '''
    TAGS = []
    ICON = None

    @classmethod
    def runner_name(cls):
        '''
        Returns the name of the Runner.
        This value does not need to be unic among your runners:
        tags are also used to differenciate them. (but you'll
        probably need to ensure (name,tags) are unique.)
        '''

        return cls.__name__
    
    @classmethod
    def runner_tags(cls):
        '''
        Returns a list of tags for this runner.
        Tags are used to group and sort runner in GUIs, and
        to uniquify runners.

        Only the first tag is used to group runners in UIs.

        Default is to return the `TAGS` class attribute.
        '''
        return cls.TAGS
    
    @classmethod
    def has_tags(cls, tags):
        '''
        Returns True if all given tags are in `runner_tags()`
        '''
        return set(tags).issubset(set(cls.runner_tags()))
    
    @classmethod
    def runner_icon(cls):
        '''
        This value depends on how you entend to use it...
        Default is to return the `ICON` class attribute.
        '''
        return cls.ICON

    @classmethod
    def can_edit(cls, filename):
        '''
        Must be implemented to return True if the filename
        looks like something supported by this process.
        
        Default is to return False
        '''
        return False

    @classmethod
    def supported_versions(cls):
        '''
        Return a list of supported version names.
        A version name of `None` represents the default version.

        Default is to return [None]
        '''
        return [None]

    def __init__(
        self,
        version=None, 
        label=None, 
        extra_argv=[],
        extra_env={}
    ):
        super(Runner, self).__init__()
        self.version = version
        self.label = label or self.runner_name()
        self.extra_argv = extra_argv
        self.extra_env = extra_env

        self._popen = None

    def show_terminal(self):
        return True

    def keep_terminal(self):
        return True

    def executable(self):
        '''
        Must return the path of the executable to run
        depending on self.version
        '''
        raise NotImplementedError()

    def argv(self):
        '''
        Must return the list of arg values for the command to run
        including self.extra_argv
        '''
        raise NotImplementedError()

    def env(self):
        '''
        Returns the env to use for the command to run.
        Default is a copy of os.environ and update with
        self.extra_env
        '''
        env = os.environ.copy()
        env.update(self.extra_env)
        return env

    def run(self):
        cmd = [self.executable()]
        cmd.extend(self.argv())

        env = self.env()
        
        os_flags = {}

        # Disowning processes in linux/mac
        if hasattr(os, 'setsid'):
            os_flags['preexec_fn'] = os.setsid

        # Disowning processes in windows
        if hasattr(subprocess, 'STARTUPINFO'):
            # Detach the process
            os_flags['creationflags'] = subprocess.CREATE_NEW_CONSOLE

            # Hide the process console
            startupinfo = subprocess.STARTUPINFO()
            if self.show_terminal():
                flag = '/C'
                if self.keep_terminal():
                    flag = '/K'
                cmd = ['cmd', flag] + cmd
            else:
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

            os_flags['startupinfo'] = startupinfo

        logger.info('Running Subprocess: %r', cmd)
        popen = subprocess.Popen(
            cmd, env=env, **os_flags
        )
        self.popen = popen

    def is_running(self):
        return self.popen is not None

    def get_log_path(self):
        return None # not yet functionnal 

class Factory(object):
    '''
    A Factory is able to find and instanciate registered Runners.
    You manage a collection of Factory with the Factories class.
    '''
    def __init__(self, name):
        super(Factory, self).__init__()
        '''
        The `name` only purpose is logging info.
        '''
        self.name = name
        self._runner_types = []

    def ensure_runner_type(self, RunnerType):
        if RunnerType in self._runner_types:
            return
        self._runner_types.append(RunnerType)

    def list_runner_types(self):
        return list(self._runner_types)

    def runner_tags(self):
        '''
        Returns an ordered list of know tags.
        '''
        tags = [] # dont use set to keep order
        for RunnerType in self._runner_types:
            for t in RunnerType.runner_tags():
                if t not in tags:
                    tags.append(t)
        return tags

    def find_runners(self, edited_filename=None, tags=[]):
        '''
        Returns a list of (runner names, tags) supporting the given
        filename and having given tags.

        An `edited_filename` of `None` will match any RunnerType
        '''
        names_and_tags = []
        for RunnerType in self._runner_types:
            if not RunnerType.has_tags(tags):
                continue
            if (
                edited_filename is None 
                or 
                RunnerType.can_edit(edited_filename)
            ):
                names_and_tags.append((
                    RunnerType.runner_name(),
                    RunnerType.runner_tags(),
                ))
        return names_and_tags

    def get_runner_versions(self, runner_name, tags=[]):
        '''
        Only the first matching runner is processed.
        If the runner is not found, None is returned.

        Using None as `tags` will match any RunnerType
        with `runner_name`
        '''
        for RunnerType in self._runner_types:
            if tags is not None and not RunnerType.has_tags(tags):
                continue
            if RunnerType.runner_name() == runner_name: 
                return RunnerType.supported_versions()
        return None

    def get_runner(
        self, 
        runner_name, tags=[], 
        version=None, label=None,
        extra_argv=[], extra_env={},
    ):
        '''
        Returns the first runner with the given runner_name and tags,
        configured with version, label, extra_arv and extra_env.
        '''
        for RunnerType in self._runner_types:
            if RunnerType.runner_name() != runner_name:
                continue
            if not RunnerType.has_tags(tags):
                continue
            return RunnerType(
                version=version, 
                label=label,
                extra_argv=extra_argv, 
                extra_env=extra_env,
            )
        return None

class Factories(object):
    
    # NB: I don't use the composition pattern here
    # because I don't want to support a tree of 
    # factories.

    def __init__(self):
        super(Factories, self).__init__()
        self._factories = []

    def create_new_factory(self, factory_name):
        '''
        Will return a new factory, but WILL NOT REGISTER it.
        '''
        return Factory(factory_name)

    def ensure_factory(self, factory):
        if factory in self._factories:
            return
        self._factories.append(factory)

    def list_runner_types(self):
        '''
        Returns an ordered list of:
            (factory, RunnerType)
        '''
        RunnerTypes = []
        for factory in self._factories:
            for RunnerType in factory.list_runner_types():
                RunnerTypes.append((factory, RunnerType))
        return RunnerTypes

    def runner_tags(self):
        tags = [] # dont use set to keep order
        for factory in self._factories:
            for tag in factory.runner_tags():
                if tag not in tags:
                    tags.append(tag)
        return tags

    def find_runners(self, edited_filename, tags=[]):
        names_and_tags = [] # dont use set to keep order
        for factory in self._factories:
            for nat in factory.find_runners(
                edited_filename, tags
            ):
                names_and_tags.append(nat)
        return names_and_tags

    def get_runner_versions(self, runner_name, tags=[]):
        '''
        Only the first matching runner is processed.
        If the runner is not found, None is returned.
        '''
        for factory in self._factories:
            versions = factory.get_runner_versions(
                runner_name, tags, 
            )
            if versions is not None:
                return versions
        return None

    def get_runner(
        self, 
        runner_name, tags=[], 
        version=None, label=None,
        extra_argv=[], extra_env={},
    ):
        '''
        Only the first matching runner is processed.
        '''
        for factory in self._factories:
            runner = factory.get_runner(
                runner_name, tags, 
                version, label,
                extra_argv, extra_env,
            )
            if runner is not None:
                return runner
        return None

class SubprocessManager(object):
    '''
    The SubprocessManager manages a list of Runner instances.
    '''

    def __init__(self):
        super(SubprocessManager, self).__init__()
        self._runners = []
    
    def get_runner_infos(self):
        '''
        Return a list of dict with keys:
            label, name, icon, version, is_running, log_path
        '''
        infos = []
        for runner in self._runners:
            infos.append(dict(
                label=runner.label,
                name=runner.runner_name(),
                icon=runner.runner_icon(),
                version=runner.version,
                is_running=runner.is_running(),
                log_path=runner.get_log_path(),
            ))

    def run(self, runner):
        self._runners.append(runner)
        runner.run()


def test():

    class MyRunner1(Runner):
        TAGS = ['A', 'B', 'C', 'D']
        
    class MyRunner2(Runner):
        TAGS = ['C', 'D', 'E', 'F']
        
    factory_1 = Factory('F1')
    factory_1.ensure_runner_type(MyRunner1)

    factory_2 = Factory('F2')
    factory_2.ensure_runner_type(MyRunner2)

    factories = Factories()
    factories.ensure_factory(factory_1)
    factories.ensure_factory(factory_2)

    tags = factories.runner_tags()
    print('TAGS:', tags)
    assert(tags == ['A', 'B', 'C', 'D', 'E', 'F'])

    runners = factories.find_runners(None, ['D', 'C'])
    print('Runners:', runners)
    assert(runners == [('MyRunner1', ['A', 'B', 'C', 'D']), ('MyRunner2', ['C', 'D', 'E', 'F'])])

if __name__ == '__main__':
    test()



from kabaret.app import resources

from kabaret.app.ui.gui.widgets.widget_view import (
    DockedView,
    QtWidgets, QtGui, QtCore
)

class SubprocessView(DockedView):

    def __init__(self, *args, **kwargs):
        super(SubprocessView, self).__init__(*args, **kwargs)

    def _build(
            self,
            top_parent, top_layout,
            main_parent,
            header_parent, header_layout
    ):
        label = QtWidgets.QLabel(main_parent)
        label.setText(
'''
<h1> Not Yet Implemented...</h1>
In an awesome future, you will see here a list of all processes running along
with their logged output and a few action to manage them.
<br>
In the meantime, enjoy the zen of python ;)
'''
        )

        lo = QtWidgets.QVBoxLayout()
        lo.addWidget(label)
        lo.addStretch(100)
        main_parent.setLayout(lo)

    def receive_event(self, event_type, data):
        pass





from kabaret import flow


class RunAction(flow.Action):

    def runner_name_and_tags(self):
        '''
        Must be implemented to return the name and the tags identifying the 
        runner to launch with this action:
            'Blender', ['Previz']

        Using None for tags will skip tag filtering and match
        the first runner with the name.
        '''
        raise NotImplementedError()

    def extra_argv(self):
        '''
        Returns the extra command line arguments to
        pass to the subprocess.

        Default is to return []
        '''
        return []
    
    def extra_env(self):
        '''
        Returns the extra env vars to
        pass to the subprocess.

        Default is to return {}
        '''
        return {}

    def get_run_label(self):
        '''
        Must return the label of the subprocess
        is the SubprocessManager UIs.

        Default is to return None so that the 
        Runner name is used as label.
        '''
        return None

    def get_versions(self):
        name, tags = self.runner_name_and_tags()
        return self.root().session().cmds.SubprocessManager.get_runner_versions(
            name, tags
        )
        
    def get_version(self, button):
        '''
        If the launched version depends on the clicked button,
        here is the chance to choose it.
        Default is to return None wich always points to the
        defalut version.
        '''
        return None
    
    def run(self, button):
        name, tags = self.runner_name_and_tags()
        self.root().session().cmds.SubprocessManager.run(
            runner_name=name,
            tags=tags,
            version=self.get_version(button),
            label=self.get_run_label(),
            extra_argv=self.extra_argv(),
            extra_env=self.extra_env(),
        )


import os
import sys
from setuptools import setup, find_packages


# (
#    (Major, Minor, [Micros]),
#    [(releaselevel, serial)],
# )
#__version_info__ = ((1, 0, 0),())
#__version_info__ = ((1, 0, 0),(rc,))
#__version_info__ = ((1, 0, 0),(rc,1))
__version_info__ = ((1, 0, 1), ())


readme = os.path.normpath(os.path.join(__file__, '..', 'README.md'))
with open(readme, 'r') as fh:
    long_description = fh.read()

# long_description += '\n\n'
# changelog = os.path.normpath(os.path.join(__file__, '..', 'CHANGELOG.md'))
# with open(changelog, 'r') as fh:
#     long_description += fh.read()


def get_version():
    global __version_info__
    return (
        '.'.join(str(x) for x in __version_info__[0])
        + ''.join(str(x) for x in __version_info__[1])
    )


setup(
    name='kabaret.subprocess_manager',
    version=get_version(),
    description='Subprocess Management extension for the Kabaret framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/kabaretstudio/kabaret.subprocess_manager',
    author='Damien "dee" Coureau',
    author_email='kabaret-dev@googlegroups.com',
    license='LGPLv3+',
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',

        'Topic :: System :: Shells',

        'Intended Audience :: Developers',
        # 'Intended Audience :: End Users/Desktop',

        'Operating System :: Microsoft :: Windows :: Windows 10',

        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',

        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    keywords='kabaret vfx animation pipeline dataflow workflow subprocess',
    install_requires=[
        'kabaret>=2.1.0b1',
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*',
    packages=find_packages('python'),
    package_dir={'': 'python'},
    package_data={
        '': ['*.css', '*.png', '*.svg', '*.gif'],
    },
)
